/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DepthEdgeFeatures.hxx
 *  Implementation of the parameter class DepthEdgeFeatures.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 21.02.2013
 */

#ifndef _DEPTHEDGEFEATURES_HXX_
#define _DEPTHEDGEFEATURES_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/DepthEdgeFeatures.h>

#include <charon-utils/ImgTool.hxx>
#include <hekate/openCvUtil.hxx>
#include <cmath>
#include <iostream>
#include <iomanip>

#include <ctime>

template <typename T>
DepthEdgeFeatures<T>::DepthEdgeFeatures(const std::string& name) :
		TemplatedParameteredObject<T>(
			"DepthEdgeFeatures", name,
			"<h2>Calculate different feature images used for the depth edge learn algorithm</h2><br>"
			"This module provides a function which calculates different feature images of a image sequence. "
			"The function has to be accessed by the given output pointer which refers to the module itself. "
		)
{
	
	ParameteredObject::_addInputSlot(
		images, "images",
		"Path to image sequence for feature generation",
		"CImgList<T>");

	ParameteredObject::_addParameter(filename, "filename", "Path and filename pattern",
			"filewrite");
	
	ParameteredObject::_addParameter< int >(
		startFrame, "startFrame", "Id of the first image in sequence.", 0);
	
	ParameteredObject::_addParameter< bool >(
		ignorePattern, "ignorePattern", "Ignore pattern to save file.", true);
	
	ParameteredObject::_addParameter< bool >(
		useOriginal, "useOriginal", "Use original source image for feature calculation.", true);
	
	ParameteredObject::_addParameter< bool >(
		useNormalized, "useNormalized", "Use normalized source image for feature calculation.", false);
	
	ParameteredObject::_addParameter< bool >(
		useEqualized, "useEqualized", "Use equalized source image for feature calculation.", false);

	ParameteredObject::_addParameter< float >(
		sigma1, "sigma1", "Sigma for gaussian blur of color image.", 2);
	ParameteredObject::_addParameter< float >(
		sigma2, "sigma2", "Sigma for gaussian blur of scale levels	.", 2);
	
	ParameteredObject::_addParameter< bool >(
		useRGB, "useRGB", "Use RGB values as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useRGBGradient, "useRGBGradient", "Use gradient as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useRGBGradientHistogram, "useRGBGradientHistogram", "Use gradient x and y histogram as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useRGBGradientHistogramXY, "useRGBGradientHistogramXY", "Use gradient sqrt(xx+yy) histogram as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesRGB, "useEigenValuesRGB", "Use Eigen Values of structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesRGBHistogram, "useEigenValuesRGBHistogram", "Use Histogram of Eigen Values of structure tensor", false);
	/*
	ParameteredObject::_addParameter< bool >(
		useEigenVectorsRGB, "useEigenVectorsRGB", "Use Eigen Vectors of structure tensor", false);
	*/
	ParameteredObject::_addParameter< bool >(
		useEigenValueRatioRGB, "useEigenValueRatioRGB", "Use ratio between Eigen Values of structure tensor", true);
	/*
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesRGB, "useEdgeAnglesRGB", "Use angles between edges and Y-axis", false);
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesDiffRGB, "useEdgeAnglesDiffRGB", "Use Difference between edge angles of neighbouring pixels", false);
	*/
	ParameteredObject::_addParameter< bool >(
		useNeighbourEigenPropertiesRGB, "useNeighbourEigenPropertiesRGB", 
		"Apply all selected Eigen features to neighbour pixels", true);
	ParameteredObject::_addParameter< bool >(
		useNeighbourRGB, "useNeighbourRGB", "Use neighbour RGB as feature.", true);
	
	ParameteredObject::_addParameter< bool >(
		useHSV_H, "useHSV_H", "Use HSV hue values as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useHSV_S, "useHSV_S", "Use HSV saturation values as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useHSV_V, "useHSV_V", "Use HSV brightness values as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useColorHistogram, "useHSVHistogram", "Use histogram of neighbour hues as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useHSVGradientH, "useHSVGradientH", "Use gradient of hue ( HSV image type ) as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useHSVGradientS, "useHSVGradientS", "Use gradient of saturation ( HSV image type ) as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useHSVGradientV, "useHSVGradientV", "Use gradient of brightness ( HSV image type ) as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useHSVGradientHistogramH, "useHSVGradientHistogramH", "Use gradient of hue histogram as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useHSVGradientHistogramS, "useHSVGradientHistogramS", "Use gradient of saturation histogram as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useHSVGradientHistogramV, "useHSVGradientHistogramV", "Use gradient of brightness histogram as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesHSV_H, "useEigenValuesHSV_H", "Use Eigen Values of hue structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesHSV_S, "useEigenValuesHSV_S", "Use Eigen Values of saturation structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesHSV_V, "useEigenValuesHSV_V", "Use Eigen Values of brightness structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesHSVHistogram_H, "useEigenValuesHSVHistogram_H", "Use Histogram of Eigen Values of hue structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesHSVHistogram_S, "useEigenValuesHSVHistogram_S", "Use Histogram of Eigen Values of saturation structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesHSVHistogram_V, "useEigenValuesHSVHistogram_V", "Use Histogram of Eigen Values of brightness structure tensor", false);
	/*
	ParameteredObject::_addParameter< bool >(
		useEigenVectorsHSV_H, "useEigenVectorsHSV_H", "Use Eigen Vectors of hue structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenVectorsHSV_S, "useEigenVectorsHSV_S", "Use Eigen Vectors of saturation structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenVectorsHSV_V, "useEigenVectorsHSV_V", "Use Eigen Vectors of bightness structure tensor", false);
	*/
	ParameteredObject::_addParameter< bool >(
		useEigenValueRatioHSV_H, "useEigenValueRatioHSV_H", "Use ratio between Eigen Values of hue structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValueRatioHSV_S, "useEigenValueRatioHSV_S", "Use ratio between Eigen Values of saturation structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValueRatioHSV_V, "useEigenValueRatioHSV_V", "Use ratio between Eigen Values of bightness structure tensor", false);
	/*
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesHSV_H, "useEdgeAnglesHSV_H", "Use angles between edges and Y-axis", false);
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesHSV_S, "useEdgeAnglesHSV_S", "Use angles between edges and Y-axis", false);
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesHSV_V, "useEdgeAnglesHSV_V", "Use angles between edges and Y-axis", false);
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesDiffHSV_H, "useEdgeAnglesDiffHSV_H", "Use Difference between edge angles of neighbouring pixels", false);
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesDiffHSV_S, "useEdgeAnglesDiffHSV_S", "Use Difference between edge angles of neighbouring pixels", false);
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesDiffHSV_V, "useEdgeAnglesDiffHSV_V", "Use Difference between edge angles of neighbouring pixels", false);
	*/
	ParameteredObject::_addParameter< bool >(
		useNeighbourEigenPropertiesHSV_H, "useNeighbourEigenPropertiesHSV_H", 
		"Apply all selected Eigen features to neighbour pixels", false);
	ParameteredObject::_addParameter< bool >(
		useNeighbourEigenPropertiesHSV_S, "useNeighbourEigenPropertiesHSV_S", 
		"Apply all selected Eigen features to neighbour pixels", false);
	ParameteredObject::_addParameter< bool >(
		useNeighbourEigenPropertiesHSV_V, "useNeighbourEigenPropertiesHSV_V", 
		"Apply all selected Eigen features to neighbour pixels", false);
	ParameteredObject::_addParameter< bool >(
		useNeighbourHSV_H, "useNeighbourHSV_H", "Use neighbour HSV as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useNeighbourHSV_S, "useNeighbourHSV_S", "Use neighbour HSV as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useNeighbourHSV_V, "useNeighbourHSV_V", "Use neighbour HSV as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useHessian, "useHessian", "Use hessian as feature", false);
	ParameteredObject::_addParameter< bool >(
		useDiffusion, "useDiffusion", "Use diffusion as feature", false);
	ParameteredObject::_addParameter< bool >(
		useBilateral, "useBilateral", "Use bilateral filter as feature", false);
	
	ParameteredObject::_addParameter< unsigned int >(
		histogramNearesNeighbours, "histogramNearesNeighbours", "Number of pixels in each direction taken into account for histograms. "
		"For example: histogramNearesNeighbours=1 -> histogram patch is 3x3, histogramNearesNeighbours=2 -> patch = 5x5", 2);
	ParameteredObject::_addParameter< unsigned int >(
		eigenvalueHistogramBins, "eigenvalueHistogramBins", "Number of eigen value histogram bins hue", 5);
	ParameteredObject::_addParameter< unsigned int >(
		gradientHistogramBinsHSV_H, "gradientHistogramBinsHSV_H", "Number of gradient of hue histogram bins. "
		"Only central bin will be used as feature.", 19);
	ParameteredObject::_addParameter< unsigned int >(
		gradientHistogramBinsHSV_S, "gradientHistogramBinsHSV_S", "Number of gradient of saturation histogram bins"
		"Only central bin will be used as feature.", 19);
	ParameteredObject::_addParameter< unsigned int >(
		gradientHistogramBinsHSV_V, "gradientHistogramBinsHSV_V", "Number of gradient of brightness histogram bins"
		"Only central bin will be used as feature.", 19);
	ParameteredObject::_addParameter< unsigned int >(
		gradientHistogramBinsRGB, "gradientHistogramBinsRGB", "Number of gradient of RGB histogram bins"
		"Only central bin will be used as feature.", 19);
	ParameteredObject::_addParameter< unsigned int >(
		gradientHistogramBinsRGBXY, "gradientHistogramBinsRGBXY", "Number of absolute gradient of RGB histogram bins", 3);
	ParameteredObject::_addParameter< unsigned int >(
		colorHistogramBinsHue, "colorHistogramBinsHue", "Number of histogram bins for hue", 6);
	ParameteredObject::_addParameter< unsigned int >(
		colorHistogramBins, "colorHistogramBins", "Number of histogram bins", 3);
	ParameteredObject::_addParameter< float >(
		neighbourDistance, "neighbourDistance", "Distance of used neighbour pixels", 1.0);
	ParameteredObject::_addParameter< float >(
		neighbourDistanceAngDiff, "neighbourDistanceAngDiff", "Distance of used neighbour pixels for angle differences", 0.5);
	ParameteredObject::_addParameter< unsigned int >(
		angDiffNearestNeighbours, "angDiffNearestNeighbours", "Number of nearest neighbours for angle differences", 6);
	
	ParameteredObject::_addParameter< int >(
		numberOfScaleSpaces, "numberOfScaleSpaces", "Number of different scale levels", 1);
	
	ParameteredObject::_setTags("Hekate;Edges") ;

}

template <typename T>
void DepthEdgeFeatures<T>::execute() {

	// Check dimensions of input slots and define parameters
	init();

	for( size_t i = 0; i < this->images().size(); i++ )
		saveFeatures( i );

}


// Calculate features and append them to feature images
template <typename T>
void DepthEdgeFeatures<T>::saveFeatures( const unsigned int index ){

	// List of feature images
	CImgList<float> features;
	
	const CImg<float> &img = this->images()[index];
	CImg<float> img_normalized, img_equalized;
	if( this->useNormalized() )
		img_normalized = this->images()[index].get_normalize(0,255);
	if( this->useEqualized() )
		img_equalized = this->images()[index].get_equalize(256);

	const unsigned int w = img.width();
	const unsigned int h = img.height();
	const unsigned int d = img.depth();
	
	// Number of different scale levels
	unsigned int levels = std::max( int(1), this->numberOfScaleSpaces() );
	sout << "Count of pyramid levels: " << levels << std::endl;

	// There are different types of images to use for calculation:
	// - original
	// - normalized 
	// - equalized
	// One iteration for each of them

	for( int i_type = 0; i_type < 3; i_type++ ){

		if( i_type == 0 && !this->useOriginal() && !this->useNormalized() && !this->useEqualized()  )
		{}
		else if( i_type == 0 && !this->useOriginal() ){
			continue;
		}
		else if( i_type == 1 && !this->useNormalized() ){
			continue;
		}
		else if( i_type == 2 && !this->useEqualized() ){
			continue;
		}
			

	// If pyramid level > 1 --> iterate over scaled images
	for( size_t i = 0; i < levels; i++ )
	{

		// Scale training image
		float blur = this->sigma1();
		if( i != 0)
			blur = pow( this->sigma2(), float(i) );
		sout << "Calculate features on blurred image with stddev of " << blur << std::endl;
		
		const CImg<float> trainScaled = img.get_blur( blur ); 
		const CImg<float> trainScaledGrey = getGreyScale( trainScaled );
		CImg<float> trainScaledHSV;
		if( img.spectrum() != 3 )
			sout << "No RGB (3) channels given" << std::endl;
		else 
			trainScaledHSV = trainScaled.get_RGBtoHSV();


		// Use Bilateral
		if(this->useBilateral())
		{
			if( img.spectrum() == 3 ){
				sout << "\tCalculate bilateral filter features on HSV" ;
				const clock_t begin_time = clock();

				CImg<float> bilH = trainScaledHSV.get_channel(0).blur_bilateral(blur+0.5, 0.2, -33, 32);
				CImg<float> bilS = trainScaledHSV.get_channel(1).blur_bilateral(blur+0.5, 0.2, -33, 32);
				CImg<float> bilV = trainScaledHSV.get_channel(2).blur_bilateral(blur+0.5, 0.2, -33, 32);
				features.push_back( bilH.get_channel(0) );
				features.push_back( bilS.get_channel(0) );
				features.push_back( bilV.get_channel(0) );
				sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
			}
			else{
				sout << "\tCalculate bilateral filter features on Grey values" ;
				const clock_t begin_time = clock();
				CImg<float> bil = trainScaled.get_channel(0).blur_bilateral(blur+0.5, 0.2, -33, 32);
				features.push_back( bil.get_channel(0) );
				sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
			}
		}

		// Use diffusion
		if(this->useDiffusion())
		{
			if( img.spectrum() == 3 ){
				sout << "\tCalculate diffusion features on HSV";
				const clock_t begin_time = clock();

				const float  sharpness = 0.7;
				const float  anisotropy = 0.6;
				const float  alpha = 0.6 * ( 1 + blur );
				const float  sigma = 1.1 * ( 1 + blur );
				CImg<float> diffH = trainScaledHSV.get_channel(0).edge_tensors(sharpness,anisotropy,alpha,sigma);
				CImg<float> diffS = trainScaledHSV.get_channel(1).edge_tensors(sharpness,anisotropy,alpha,sigma);
				CImg<float> diffV = trainScaledHSV.get_channel(2).edge_tensors(sharpness,anisotropy,alpha,sigma);
				CImg<float> diffHa = diffH.get_channel(0).abs() + diffH.get_channel(1).abs() + diffH.get_channel(2).abs() ;
				CImg<float> diffSa = diffS.get_channel(0).abs() + diffS.get_channel(1).abs() + diffS.get_channel(2).abs() ;
				CImg<float> diffVa = diffV.get_channel(0).abs() + diffV.get_channel(1).abs() + diffV.get_channel(2).abs() ;
				features.push_back( diffHa.get_channel(0) );
				features.push_back( diffSa.get_channel(0) );
				features.push_back( diffVa.get_channel(0) );			
				sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
			}
			else{
				sout << "\tCalculate diffusion features on Grey";
				const clock_t begin_time = clock();
				const float  sharpness = 0.7;
				const float  anisotropy = 0.6;
				const float  alpha = 0.6 * ( 1 + blur );
				const float  sigma = 1.1 * ( 1 + blur );
				CImg<float> diff = trainScaled.get_channel(0).edge_tensors(sharpness,anisotropy,alpha,sigma);
				features.push_back( diff.get_channel(0) );		
				sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
			}
		}

		
		// RGB image
		if(this->useRGB()){
			sout << "\tCalculate RGB features" ;
			const clock_t begin_time = clock();

			features.push_back( trainScaled.get_channel(0) );
			features.push_back( trainScaled.get_channel(1) );
			features.push_back( trainScaled.get_channel(2) );
			
			sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
		}
	
		// HSV image
		if( this->useHSV_H() || this->useHSV_S() ||this->useHSV_V() ){
			sout << "\tCalculate HSV features" ;
			const clock_t begin_time = clock();

			/*
			HSV encoding:
			H = [0,360) / S = [0,1] / V = [0,1]
			*/

			if( this->useHSV_H() )
			{
				features.push_back( trainScaledHSV.get_channel(0)  );
			}
			if( this->useHSV_S() ) 
			{
				features.push_back( trainScaledHSV.get_channel(1) );
			}
			if( this->useHSV_V() ) 
			{
				features.push_back( trainScaledHSV.get_channel(2) );
			}
			
			sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;			
		}

		// Histogram of gradients
		if((this->useRGBGradientHistogram() || this->useRGBGradientHistogramXY()) && i == 0)
		{
			sout << "\tCalculate RGB gradient histogram features" ;
			const clock_t begin_time = clock();

			// calculate number of pixels to crop
			const unsigned int size = this->histogramNearesNeighbours();
			cimg_library::CImgList<float> gradient = trainScaledGrey.get_gradient();
			
			// maximum gradient can be -255, minimum 255;
			if( gradient[0].max() > 255. )
				throw( std::runtime_error("useRGBGradientHistogram: maximum gradient > 255 ") );
			if( gradient[0].min() < -255. )
				throw( std::runtime_error("useRGBGradientHistogram: minimum gradient < -255 ") );
			if( gradient[1].max() > 255. )
				throw( std::runtime_error("useRGBGradientHistogram: maximum gradient > 255 ") );
			if( gradient[1].min() < -255. )
				throw( std::runtime_error("useRGBGradientHistogram: minimum gradient < -255 ") );
						
			// histogram of gradients in x and y direction
			if(this->useRGBGradientHistogram())
			{
			
				unsigned int bins = this->gradientHistogramBinsRGB();
				// Check if value is evan
				if( bins % 2 == 0) bins++;
				size_t centralBin = int(bins / 2);

				int binWidth = 511 / 2;
				CImg<float> gradXCut = gradient[0].get_min(binWidth).max(-binWidth);
				CImg<float> gradYCut = gradient[1].get_min(binWidth).max(-binWidth);

				// Create histogram image (width,height,numberOfBins);
				cimg_library::CImg<float> histogramX(w,h,3,1);
				cimg_library::CImg<float> histogramY(w,h,3,1);
				
				#pragma omp parallel for
				for( int xy = 0; xy < w*h; xy++ )
				{
					
					const unsigned int x = xy % w;
					const unsigned int y = xy / w;

					CImg<float> histX = gradXCut.get_crop(x-size,y-size,x+size,y+size).histogram(bins,-255.,255.);
					CImg<float> histY = gradYCut.get_crop(x-size,y-size,x+size,y+size).histogram(bins,-255.,255.);
					// Only save central bin

					histogramX(x,y,0,0) = histX(centralBin,0);
					histogramY(x,y,0,0) = histY(centralBin,0);
					histogramX(x,y,1,0) = histX(centralBin-1,0);
					histogramY(x,y,1,0) = histY(centralBin-1,0);
					histogramX(x,y,2,0) = histX(centralBin+1,0);
					histogramY(x,y,2,0) = histY(centralBin+1,0);
				}
				features.push_back( histogramX );
				features.push_back( histogramY );
			}

			// historgram of sqrt( gradX + gradY )
			if(this->useRGBGradientHistogramXY())
			{

				unsigned int binsXY = this->gradientHistogramBinsRGBXY();

				cimg_library::CImg<float> gradXY = (gradient[0].get_sqr()+gradient[1].get_sqr()).sqrt();
				// Create histogram image (width,height,numberOfBins);
				cimg_library::CImg<float> histogramXY(w,h,binsXY,1);

				#pragma omp parallel for
				for( int xy = 0; xy < w*h; xy++ )
				{					
					const unsigned int x = xy % w;
					const unsigned int y = xy / w;

					CImg<float> histXY = gradXY.get_crop(x-size,y-size,x+size,y+size).histogram(binsXY,0,std::sqrt(360.7));
					for(size_t j = 0; j < binsXY; j++)
					{
						histogramXY(x,y,j,0) = histXY(j,0);
					}
				}

				for(size_t j = 0; j < binsXY; j++)
				{
					features.push_back( histogramXY.get_slice(j) );
				}

			}
			
			sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
		}

		// Histogram of Hue (HSV)
		if(this->useColorHistogram() && i==0)
		{
			if( img.spectrum() == 3 ){
			sout << "\tCalculate HSV histogram features" ;
			const clock_t begin_time = clock();
			
			const unsigned int binsHue = this->colorHistogramBinsHue();
			const unsigned int bins = this->colorHistogramBins();
			// calculate number of pixels to crop
			const unsigned int size = this->histogramNearesNeighbours();
			// Get hue channel of HSV image
			cimg_library::CImg<float> hue = trainScaledHSV.get_channel(0);
			cimg_library::CImg<float> sat = trainScaledHSV.get_channel(1);
			cimg_library::CImg<float> val = trainScaledHSV.get_channel(2);
			// Create histogram image (width,height,numberOfBins);
			cimg_library::CImg<float> histogram(w,h,bins,3);

			#pragma omp parallel for
			for( int xy = 0; xy < w*h; xy++ )
			{					
				const unsigned int x = xy % w;
				const unsigned int y = xy / w;

				CImg<float> hist = hue.get_crop(x-size,y-size,x+size,y+size).histogram(binsHue,0,360);
				CImg<float> hist2 = sat.get_crop(x-size,y-size,x+size,y+size).histogram(bins,0,1);
				CImg<float> hist3 = val.get_crop(x-size,y-size,x+size,y+size).histogram(bins,0,1);
				for(size_t j = 0; j < bins; j++)
				{
					histogram(x,y,j,0) = hist(j,0);
					histogram(x,y,j,1) = hist2(j,0);
					histogram(x,y,j,2) = hist3(j,0);
				}

			}
			for(size_t j = 0; j < bins; j++)
			{
				features.push_back( histogram.get_slice(j).channel(0) );
				features.push_back( histogram.get_slice(j).channel(1) );
				features.push_back( histogram.get_slice(j).channel(2) );
			}
			
			sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
			}
		}

		// Gradient images
		if(this->useRGBGradient()){
			sout << "\tCalculate RGB gradient features" ;
			const clock_t begin_time = clock();

			// Gradient of train image
			cimg_library::CImgList<float> gradient = trainScaledGrey.get_gradient();
			features.push_back( gradient[0] );
			features.push_back( gradient[1] );
			
			sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
		}

		// Gradient of HSV values
		if( this->useHSVGradientH() || this->useHSVGradientS() || this->useHSVGradientV()
				|| this->useHSVGradientHistogramH() || this->useHSVGradientHistogramS() || this->useHSVGradientHistogramV() )
		{
			sout << "\tCalculate HSV gradient features" ;
			const clock_t begin_time = clock();

			// Calculate HSV gradients
			CImgList<float> gradients = getHSVGradients( trainScaledHSV );

			// Save gradients
			if( this->useHSVGradientH() )
			{
				features.push_back( gradients[0].get_channel(0) );
				features.push_back( gradients[1].get_channel(0) );
			}
			if( this->useHSVGradientS() )
			{
				features.push_back( gradients[0].get_channel(1) );
				features.push_back( gradients[1].get_channel(1) );
			}
			if( this->useHSVGradientV() )
			{
				features.push_back( gradients[0].get_channel(2) );
				features.push_back( gradients[1].get_channel(2) );
			}
			
			sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;

			// Calculate gradient histograms
			if( this->useHSVGradientHistogramH() || this->useHSVGradientHistogramS() || this->useHSVGradientHistogramV() )
			{
				sout << "\tCalculate HSV gradient histogram features" ;
				const clock_t begin_time2 = clock();

				// calculate number of pixels to crop
				const unsigned int size = this->histogramNearesNeighbours();

				// Iterate over H, S and V
				for( size_t k = 0; k < 3; k++ )
				{
					
					if( !this->useHSVGradientHistogramH() && k == 0)
						continue;
					if( !this->useHSVGradientHistogramS() && k == 1)
						continue;
					if( !this->useHSVGradientHistogramV() && k == 2)
						continue;

					// get min max values of pixels for histogram
					float min_value, max_value;
					if( this->useHSVGradientHistogramH() && k == 0)
					{
						min_value = -180.;
						max_value = 180.;
					}
					if( this->useHSVGradientHistogramS() && k == 1)
					{
						min_value = -1.;
						max_value = 1.;
					}
					if( this->useHSVGradientHistogramV() && k == 2)
					{
						min_value = -1.;
						max_value = 1.;
					}

					unsigned int bins = this->gradientHistogramBinsHSV_H();
					if(k==1) bins = this->gradientHistogramBinsHSV_S();
					if(k==2) bins = this->gradientHistogramBinsHSV_V();
					
					// Check if value is evan
					if( bins % 2 == 0) bins++;
					size_t centralBin = int(bins / 2);

					// Create histogram image (width,height,numberOfBins);
					cimg_library::CImg<float> histogramX(w,h,3,1);
					cimg_library::CImg<float> histogramY(w,h,3,1);


					int binWidth = std::abs(max_value-min_value) / 2;
					CImg<float> gradXCut = gradients[0].get_min(binWidth).max(-binWidth);
					CImg<float> gradYCut = gradients[1].get_min(binWidth).max(-binWidth);
									
					#pragma omp parallel for
					for( int xy = 0; xy < w*h; xy++ )
					{					
						const unsigned int x = xy % w;
						const unsigned int y = xy / w;

						const CImg<float> histX = gradXCut.get_crop(x-size,y-size,x+size,y+size).channel(k).histogram(bins,min_value,max_value);
						const CImg<float> histY = gradYCut.get_crop(x-size,y-size,x+size,y+size).channel(k).histogram(bins,min_value,max_value);
						// save central 3 bins
						histogramX(x,y,0,0) = histX(centralBin,0);
						histogramY(x,y,0,0) = histY(centralBin,0);
						histogramX(x,y,1,0) = histX(centralBin-1,0);
						histogramY(x,y,1,0) = histY(centralBin-1,0);
						histogramX(x,y,2,0) = histX(centralBin+1,0);
						histogramY(x,y,2,0) = histY(centralBin+1,0);
					}
					
					// if scale space 0, save two more bins
					if(i == 0){
						features.push_back( histogramX.get_slice(0) );
						features.push_back( histogramY.get_slice(0) );
					}
					
				}
				sout << " (" << float( clock () - begin_time2 ) / CLOCKS_PER_SEC << ")" << std::endl;
	
			}
			
		}

		// Hessian images
		if(this->useHessian()){
			sout << "\tCalculate hessian features" ;
			const clock_t begin_time = clock();

			// Gradient of train image
			cimg_library::CImgList<float> hessian = trainScaledGrey.get_hessian();
			features.push_back( hessian[0].abs() + hessian[1].abs() + hessian[2].abs() );
			
			sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
		}

		sout << "\tChecking Eigen Features" << std::endl;
		// two iterations ( HSV and RGB )
		for( size_t l=0; l<2; l++)
		{

			// Flags which activate structure tensor calculation
			bool uEVal=false, uEVec=false, uERat=false, uEAng=false, uEAngDiff=false, uENei=false, uENeiProp=false, uEValHist=false;
			
			bool uEVal_H = this->useEigenValuesHSV_H();
			bool uEVal_S = this->useEigenValuesHSV_S();
			bool uEVal_V = this->useEigenValuesHSV_V();
			bool uEVec_H = this->useEigenVectorsHSV_H();
			bool uEVec_S = this->useEigenVectorsHSV_S();
			bool uEVec_V = this->useEigenVectorsHSV_V();
			bool uERat_H = this->useEigenValueRatioHSV_H();
			bool uERat_S = this->useEigenValueRatioHSV_S();
			bool uERat_V = this->useEigenValueRatioHSV_V();
			bool uEAng_H = this->useEdgeAnglesHSV_H();
			bool uEAng_S = this->useEdgeAnglesHSV_S();
			bool uEAng_V = this->useEdgeAnglesHSV_V();
			bool uEAngDiff_H = this->useEdgeAnglesDiffHSV_H();
			bool uEAngDiff_S = this->useEdgeAnglesDiffHSV_S();
			bool uEAngDiff_V = this->useEdgeAnglesDiffHSV_V();
			bool uENei_H = this->useNeighbourHSV_H();
			bool uENei_S = this->useNeighbourHSV_S();
			bool uENei_V = this->useNeighbourHSV_V();
			bool uENeiProp_H = this->useNeighbourEigenPropertiesHSV_H();
			bool uENeiProp_S = this->useNeighbourEigenPropertiesHSV_S();
			bool uENeiProp_V = this->useNeighbourEigenPropertiesHSV_V();
			bool uEValHist_H = this->useEigenValuesHSVHistogram_H();
			bool uEValHist_S = this->useEigenValuesHSVHistogram_S();
			bool uEValHist_V = this->useEigenValuesHSVHistogram_V();

			// Set RGB flags
			if( l == 0 )
			{
				// RGB Structure Tensor features
				uEVal = this->useEigenValuesRGB();
				uEVec = this->useEigenVectorsRGB();
				uERat = this->useEigenValueRatioRGB();
				uEAng = this->useEdgeAnglesRGB();
				uEAngDiff = this->useEdgeAnglesDiffRGB();
				uENei = this->useNeighbourRGB();
				uENeiProp = this->useNeighbourEigenPropertiesRGB();
				uEValHist = this->useEigenValuesRGBHistogram();
			}
			// Set HSV flags
			else
			{
				// use eigen values
				if(uEVal_H || uEVal_S || uEVal_V)
					uEVal = true;
				else 
					uEVal = false;

				// use eigen vectors
				if(uEVec_H || uEVec_S || uEVec_V)
					uEVec = true;
				else 
					uEVec = false;

				// use eigen ratio
				if(uERat_H || uERat_S || uERat_V)
					uERat = true;
				else 
					uERat = false;

				// use eigen angles
				if(uEAng_H || uEAng_S || uEAng_V)
					uEAng = true;
				else 

				// use eigen angles differences
				if(uEAngDiff_H || uEAngDiff_S || uEAngDiff_V)
					uEAngDiff = true;
				else 
					uEAngDiff = false;

				// use HSV neighbors
				if(uENei_H || uENei_S || uENei_V)
					uENei = true;
				else 
					uENei = false;

				// use HSV eigen neighbours
				if(uENeiProp_H || uENeiProp_S || uENeiProp_V)
					uENeiProp = true;
				else 
					uENeiProp = false;

				// use HSV eigenvalue histograms
				if(uEValHist_H || uEValHist_S || uEValHist_V)
					uEValHist = true;
				else
					uEValHist = false;


			}

			if( uEVal || uEVec || uERat || uEAng || uEAngDiff || uENei || uENeiProp || uEValHist ){
				sout << "\tCalculate Eigen features" ;
				const clock_t begin_time = clock();

				if( l == 0 ) sout << " for RGB values" << std::endl;
				else sout << " for HSV values" << std::endl;
			
				const float dist = this->neighbourDistance() ;
				CImg<float> colorImage;	// color image (RBG/HSV)
				CImg<float> st;			// structure tensor (RBG/HSV)
			
				// calculate RGB / HSV structure tensor
				if( l == 0 )
				{
					st = trainScaledGrey.get_structure_tensors(0);
					colorImage = trainScaled;
				}
				else
				{
					// the structure tensor of an HSV image always has three channels
					// flags for H, S, V are checked afterwards for the push_back function
					colorImage = trainScaled.get_RGBtoHSV();
					st = getHSVStructureTensor( colorImage );
				}
			
				// get eigen entities from structure tensor
				CImg<float> eVal, eVec1, eVec2, eRat, eAng, eAngDiff;
				CImgList<float> eValHist;
				sout << "\t\tCalculate eigen values, vectors, ratios and angles" << std::endl;
				getEigenEntities(st, eVal, eVec1, eVec2, eRat, eAng);

				// Get histograms
				if( uEValHist ){
					sout << "\t\tCalculate histograms for eigen values etc." << std::endl;
					getEValHistograms( eValHist, eVal, eRat, this->eigenvalueHistogramBins(), this->histogramNearesNeighbours(), i );
				}
			
				// Get difference between edge angles of neighbour pixels
				if( uEAngDiff ){
					sout << "\t\tCalculate differences of eigen angles" << std::endl;
					eAngDiff = getEAngDiff( eAng, st, eVec1, this->angDiffNearestNeighbours(), this->neighbourDistanceAngDiff() );
				}
				
				// get eigen entities of neighbour pixels from structure tensor
				CImgList<float> eValN, eVec1N, eVec2N, eRatN, eAngN, colorN;
				if(uENei || uENeiProp){
					sout << "\t\tCalculate neighbour properties" << std::endl;
					getEigenEntitiesN(colorImage, st, dist, eVec1, eVec2, eValN, eVec1N, eVec2N, eRatN, eAngN, colorN);
				}
				
				// Save calculated features for every channel separately
				// ! Different channel results were saved in z-axis !

				sout << "\t\tSave feature images" << std::endl;
				for( unsigned int k = 0; k < st.depth(); k++ )
				{
				
					if(i==0)
					{
						if( (l==0 && uEValHist) || (l==1 && k==0 && uEValHist_H) || (l==1 && k==1 && uEValHist_S) || (l==1 && k==2 && uEValHist_V) ){
							for( size_t bins = 0; bins < eValHist.size(); bins++ )
							{
									features.push_back( eValHist[bins].get_channel(0).slice(k) );
									features.push_back( eValHist[bins].get_channel(1).slice(k) );
							}
						}							
					}
					if( (l==0 && uEVal) || (l==1 && k==0 && uEVal_H) || (l==1 && k==1 && uEVal_S) || (l==1 && k==2 && uEVal_V) ){
						features.push_back( eVal.get_channel(0).slice(k) );
					}
					if( (l==0 && uERat) || (l==1 && k==0 && uERat_H) || (l==1 && k==1 && uERat_S) || (l==1 && k==2 && uERat_V) ){
						features.push_back( eRat.get_channel(0).slice(k) );
					}
					// RGB eigen entities of neigbour pixels
					if( (l==0 && uENeiProp) || (l==1 && k==0 && uENeiProp_H) || (l==1 && k==1 && uENeiProp_S) || (l==1 && k==2 && uENeiProp_V) )
					{
						for( size_t j=0; j < 4; j++)
						{
							features.push_back( eValN[i].get_channel(0).slice(k) );
							features.push_back( eRatN[j].get_channel(0).slice(k) );
						}
					}
					// RGB/HSV of neigbour pixels in eigen vector directions
					if( (l==0 && uENei) || (l==1 && k==0 && uENei_H) || (l==1 && k==1 && uENei_S) || (l==1 && k==2 && uENei_V) )
					{
						for( size_t j=0; j < 4; j++)
						{
							features.push_back( colorN[j].get_channel(0).slice(k) );
						}
					}

				}
					
				sout << "\tTook " << float( clock () - begin_time ) / CLOCKS_PER_SEC << "seconds" << std::endl;
			
			}

		}

	}

	}

	// SAVE FILE
	writeToFile( features, index );

}

/// Save file
template <typename T>
bool DepthEdgeFeatures<T>::writeToFile( CImgList<float> &features, const int index ){
	
	std::string workingPath = FileTool::getCurrentDir();

	std::string base = this->filename();
	//if(base == "") throw(std::runtime_error("FileWriterSequence<T>::execute(): No filename given"));
	
	std::string ext =  this->filename().substr( this->filename().find_last_of('.'),  this->filename().length());
	std::string name =  this->filename().substr( 0, this->filename().find_last_of('.'));

	std::stringstream s;
	int id  = index ;
	if( !this->ignorePattern() )
		id += this->startFrame();

	s << std::setw(6) << std::setfill('0') << id;

	sout << "	" << name << "_" << s.str() << ext << std::endl;

	features.save( ( name + "_" + s.str() + ".cimg" ).c_str() );

	FileTool::changeDir( workingPath );

	return 1;

}


/// Get differencens between angles of eigenvectors of neighbour pixels in edge direction
/** Sum the difference of edge angles of non (number of nearest neighbours) neighbouring pixels
	in each direction. The distance of neighbouring pixels is defined by nndist.
	*/
CImg<float> getEAngDiff( const CImg<float> &eAng, const CImg<float> &st, const CImg<float> &eVec1, 
						const unsigned int non, const float nndist )
{

	CImg<float> diff(st.width(), st.height(), st.depth(), 1);
	diff.fill( 0 );
	
	const unsigned int w = st.width();
	const unsigned int h = st.height();

	#pragma omp parallel for
	for( int xy = 0; xy < w*h; xy++ )
	{					
		const unsigned int x = xy % w;
		const unsigned int y = xy / w;

		for(int z = 0; z < st.depth(); z++ )
		{

			// Temporary neighbour buffers
			CImg<float> vec, mat2, val2, vec2;
			mat2.assign(2,2);
			val2.assign(1,2);
			vec2.assign(2,2);
			vec.assign(2,2);
			vec.fill(0);

			vec(0,0) = eVec1(x,y,z,0);
			vec(0,1) = eVec1(x,y,z,1);

			// build differences in both directions of eigen vector
			for( int s = -1; s <= 1; s = s+2 )
			{
				float newX = x;
				float newY = y;
				vec2 = vec;
				float ang = acos(vec(0,1));

				for( size_t i = 0; i < non; i++ )
				{
					// Fill structure tensor at neighbour position (x,y)
					newX += s*i*nndist*vec2(0,0) ;
					newY += s*i*nndist*vec2(0,1) ;

					mat2.fill(
							st.linear_atXY(newX,newY,z,0),st.linear_atXY(newX,newY,z,1),
							st.linear_atXY(newX,newY,z,1),st.linear_atXY(newX,newY,z,2)
					);
					
					// calculate eigenvectors and eigenvalues
					mat2.symmetric_eigen(val2,vec2);
					
					// fill into output slots
					assert(val2.width()==1);
					assert((val2[1] <= val2[0]));
					assert(vec2.width()==vec2.height());
					assert(vec2==vec2);
					assert(val2[0]==val2[0]);
					assert(val2[1]==val2[1]);

					// Calculate angles of new vec2
					float ang2 = acos( vec2(0,1) );
				
					// Add difference
					diff(x,y,z,0) += std::min( std::abs(ang2-ang), float(PI) - std::abs(ang2-ang) );

					// Save new angle for next iteration
					ang = ang2;

				}

			}

		}

	}

	return diff;


}


// Returns grey scale image
template <typename T>
CImg<T> DepthEdgeFeatures<T>::getGreyScale( const CImg<T> &img )
{



	// Create grey scale image
	CImg<T> out( img.width(), img.height(), 1, 1 );
	out.fill( T(0) );

	if( img.spectrum() == 0 )
		throw( std::range_error("img.spectrum() == 0") );

	// sum up image channels
	for( unsigned int i = 0; i < img.spectrum(); i++)
	{
		out += img.get_channel(i) ;
	}

	// divide by number of channels
	out /= img.spectrum() ;
	
	return out;

}


/// Returns gradient for hue of HSV values
/** 
Returns a CImgList with two images. First image contains the x deviations, second the
the y deviations. The channels of the images contain the gradients of H, S, and V.
*/
template <typename T>
CImgList<float> getHSVGradients( const CImg<T> &img )
{

	if( img.depth() != 1)
		throw( std::range_error("getHSVGradient: image depth != 1") );

	cimg_library::CImgList<float> grad;
	CImg<float> gradX(img.width(), img.height(), 1, 3 );
	CImg<float> gradY(img.width(), img.height(), 1, 3 );

	// Create convolve mask
	CImg<float> cX(3,1), cY(1,3);
	cX[0] = -0.5;	cX[1] = 0;		cX[2] = 0.5;
	cY[0] = -0.5;	cY[1] = 0;		cY[2] = 0.5;
	
	gradX = img.get_convolve( cX );
	gradY = img.get_convolve( cY );

	// Correct hue values due to overlap
	const unsigned int w = img.width();
	const unsigned int h = img.height();

	#pragma omp parallel for
	for( int xy = 0; xy < w*h; xy++ )
	{					
		const unsigned int x = xy % w;
		const unsigned int y = xy / w;

		// Correct gradX (overlap between [0;369) )
		const float xl = gradX(x,y,0,0);
		const float xlSign = ( xl < 0 ? -1 : 1 );
		gradX(x,y,0,0) = xlSign * std::min( std::abs(xl), float(360.) - std::abs(xl) );
		// Correct gradY (overlap between [0;369) )
		const float yl = gradY(x,y,0,0);
		const float ylSign = ( yl < 0 ? -1 : 1 );
		gradY(x,y,0,0) = ylSign * std::min( std::abs(yl), float(360.) - std::abs(yl) );
	}
	
	grad.push_back( gradX );
	grad.push_back( gradY );

	return grad;

}


/// Returns structure tensor for HSV image type
/** 
Returns the structure tensors for an HSV image type. Attention: The three matrix entries 
are save in channel-axis while if the structure tensor is calculated on an
image with several channels, the different channel values are saved in z-axis.
z = ( H, S, V )
c = ( Ix*Ix, Ix*Iy, Iy*Iy )
*/
template <typename T>
CImg<float> getHSVStructureTensor( const CImg<T> &img )
{

	const CImgList<float> gradients = getHSVGradients( img );
	CImg<float> st(img.width(), img.height(), 3, 3);

	const unsigned int w = img.width();
	const unsigned int h = img.height();

	#pragma omp parallel for
	for( int xy = 0; xy < w*h; xy++ )
	{				
		const unsigned int x = xy % w;
		const unsigned int y = xy / w;

		for( int c = 0; c < 3; c++ )
		{
			st(x,y,c,0) = gradients[0](x,y,0,c) * gradients[0](x,y,0,c);
			st(x,y,c,1) = gradients[0](x,y,0,c) * gradients[1](x,y,0,c);
			st(x,y,c,2) = gradients[1](x,y,0,c) * gradients[1](x,y,0,c);
		}

	}
	
	return st;

}


/// Calculate eigen value histograms
template <typename T>
void DepthEdgeFeatures<T>::getEValHistograms( CImgList<float> &hist, const CImg<float> &eVal, const CImg<float> &eRat, 
											 const unsigned int bins, const unsigned int size, const unsigned int scaleSpace )
{
		
	const unsigned int w = eVal.width();
	const unsigned int h = eVal.height();
	const unsigned int d = eVal.depth();

	CImgList<float> list( bins, w, h, d, 2, 0 );

	// Iterate over number of given eigen values
	for( size_t k = 0; k < d; k++ )
	{

		CImg<float> ev = eVal.get_slice(k).channel(0);
		CImg<float> rat = eRat.get_slice(k).channel(0);

		// maximum value of eigen values in the current image.
		// Actually, the maximum of all images had to be taken
		float maxEV, maxRatio;

		if( d > 1 )
		{
			if(_maxEV_HSV[scaleSpace][k] == 0) _maxEV_HSV[scaleSpace][k] = ev.max();
			if(_maxEVRatio_HSV[scaleSpace][k] == 0) _maxEVRatio_HSV[scaleSpace][k] = rat.max();
			maxEV = _maxEV_HSV[scaleSpace][k];
			maxRatio = _maxEVRatio_HSV[scaleSpace][k];
		}
		else
		{
			if(_maxEV_RGB[scaleSpace] == 0) _maxEV_RGB[scaleSpace] = ev.max();
			if(_maxEVRatio_RGB[scaleSpace] == 0) _maxEVRatio_RGB[scaleSpace] = rat.max();
			maxEV = _maxEV_RGB[scaleSpace];
			maxRatio = _maxEVRatio_RGB[scaleSpace];
		}
		
		ev.cut(0,maxEV);
		rat.cut(0,maxRatio);

		#pragma omp parallel for
		for( int xy = 0; xy < ev.width() * ev.height(); xy++ )
		{					
			const unsigned int x = xy % w;
			const unsigned int y = xy / w;

			const CImg<float> hist1 = ev.get_crop(x-size,y-size,x+size,y+size).histogram(bins,0,maxEV);
			const CImg<float> hist2 = rat.get_crop(x-size,y-size,x+size,y+size).histogram(bins,0,maxRatio);
			for(size_t i = 0; i < bins; i++)
			{
				list[i](x,y,k,0) = hist1(i,0);
				list[i](x,y,k,1) = hist2(i,0);
			}

		}

	}

	// Only take the first two bins
	//hist = list;
	
	hist.push_back(list[1]);
	for(size_t i = 0; i < bins; i++)
	{
		hist.push_back(list[i]);
	}

}

/// Calculates eigen entities such as eigen values and vectors from structure tensor of neighbour pixels
/** The neighbour pixels are determined with the given eigen vectors of the source pixel.*/
void getEigenEntitiesN(const CImg<float> &img, const CImg<float> &st, const float dist, const CImg<float> &eVec1, 
			const CImg<float> &eVec2, CImgList<float> &eValN, CImgList<float> &eVec1N, CImgList<float> &eVec2N, 
			CImgList<float> &eRatN, CImgList<float> &eAngN, CImgList<float> &colorN){

	sout << "\t\t\t\tCalculate neighbour values" << std::endl;
				
	// dimensions of structure tensor
	const unsigned int w = st.width();
	const unsigned int h = st.height();
	const unsigned int d = st.depth();

	// Resize for 4 nearest neighbours
	for( size_t i=0; i < 4; i++)
	{
		eValN.push_back( CImg<float>(w, h, d, 2).fill(0) );
		eVec1N.push_back( CImg<float>(w, h, d, 2).fill(0) );
		eVec2N.push_back( CImg<float>(w, h, d, 2).fill(0) );
		eRatN.push_back( CImg<float>(w, h, d, 1).fill(0) );
		eAngN.push_back( CImg<float>(w, h, d, 1).fill(0) );
		colorN.push_back( CImg<float>(w, h, d, 1).fill(0) );
	}

	sout << "\t\t\t\tCalculate neighbour values 1" << std::endl;

	#pragma omp parallel for
	for( int xy = 0; xy < w*h; xy++ )
	{					
		const unsigned int x = xy % w;
		const unsigned int y = xy / w;

		for( int z=0; z < d; z++)
		{

			// Temporary neighbour buffers
			CImg<float> vec, mat2, val2, vec2;
			mat2.assign(2,2);
			val2.assign(1,2);
			vec2.assign(2,2);
			vec.assign(2,2);

			vec(0,0) = eVec1(x,y,z,0);
			vec(0,1) = eVec1(x,y,z,1);
			vec(1,0) = eVec2(x,y,z,0);
			vec(1,0) = eVec2(x,y,z,1);
		
			int k = 0;
			// iterate over four neighbour pixels
			for( int i = -1; i<2; i+=2 )
			{
				for( int j = 0; j<2; j++, k++ )
				{
					try{
						// Fill structure tensor at neighbour position (x,y)
						float newX = float(x) + float(i*dist)*vec(j,0) ;
						float newY = float(y) + float(i*dist)*vec(j,1) ;

						// Border conditions: if outside, continue
						if( newX < 0 || newY < 0 || newX > w-1 || newY > h-1 )
							continue;

						float clr = img.linear_atXY(newX,newY,0,z,std::numeric_limits<float>::quiet_NaN());
						if( clr == std::numeric_limits<float>::quiet_NaN() )
							continue;

						if ( clr!=clr )
							continue;
						
						colorN[k](x,y,z,0) = clr ;

						mat2.fill(
								st.linear_atXY(newX,newY,z,0,0),st.linear_atXY(newX,newY,z,1,0),
								st.linear_atXY(newX,newY,z,1,0), st.linear_atXY(newX,newY,z,2,0)
						);
					
						// calculate eigenvectors and eigenvalues
						mat2.symmetric_eigen(val2,vec2);
					
						// fill into output slots
						assert(val2.width()==1);
						assert((val2[1] <= val2[0]));
						assert(vec2.width()==vec2.height());
						assert(vec2==vec2);
						assert(val2[0]==val2[0]);
						if(val2[1]!=val2[1] ||val2[0]!=val2[0])
							continue;
				
						eValN[k](x,y,z,0) = val2[0];
						eValN[k](x,y,z,1) = val2[1];
						eVec1N[k](x,y,z,0) = vec2(0,0);
						eVec1N[k](x,y,z,1) = vec2(0,1);
						eVec2N[k](x,y,z,0) = vec2(1,0);
						eVec2N[k](x,y,z,1) = vec2(1,1);	

						eRatN[k](x,y,z,0) = val2[1]*val2[1]+val2[0]*val2[0] ;

						eAngN[k](x,y,z,0) = acos( vec2(0,1) );
					
						// calculate colors of both sides of the gradient
						// cimg_library::cubic_atXY is not stable against OpenMP
						//int cropx = int(newX), cropy = int(newY);
						//CImg<float> crop = img.get_crop(newX-1,newY-1,newY+1,newY+1);

					}
					catch(...){
						continue;
					}	

				}

			}

		}
				
	}
	
}

/// Calculates eigen entities such as eigen values and vectors from structure tensor
void getEigenEntities(const CImg<float> &st, CImg<float> &eVal, CImg<float> &eVec1, 
			CImg<float> &eVec2, CImg<float> &eRat, CImg<float> &eAng){
				
	const unsigned int w = st.width();
	const unsigned int h = st.height();
	const unsigned int d = st.depth();

	// Eigen values separated in channel 0 and 1
	eVal.resize( w, h, d, 2 );

	// Eigen vectors. Coordinates separated in channels 0 and 1. (channel0 =x, channel1=y)
	eVec1.resize(w, h, d, 2);
	eVec2.resize(w, h, d, 2);

	// Ratio between eigen values and angles between eigen vectors
	eRat.resize(w, h, d, 1);
	eAng.resize(w, h, d, 1);		

	// Calculate values for every pixel
	#pragma omp parallel for
	for( int xy = 0; xy < w*h; xy++ )
	{

		// Temporary buffers
		CImg<float> mat, val, vec;
		mat.assign(2,2);
		val.assign(1,2);
		vec.assign(2,2);	

		const unsigned int x = xy % w;
		const unsigned int y = xy / w;

		for( int z = 0; z < st.depth(); z++)
		{

			// Fill structure tensor at pixel position (x,y)
			mat.fill(
					st(x,y,z,0),st(x,y,z,1),
					st(x,y,z,1),st(x,y,z,2)
			);

			// calculate eigenvectors and eigenvalues
			mat.symmetric_eigen(val,vec);

			// fill into output slots
			assert(val.width()==1);
			assert((val[1] <= val[0]));
			assert(vec.width()==vec.height());
			assert(vec==vec);
			assert(val[0]==val[0]);
			assert(val[1]==val[1]);
				
			eVal(x,y,z,0) = val[0];
			eVal(x,y,z,1) = val[1];
			eVec1(x,y,z,0) = vec(0,0);
			eVec1(x,y,z,1) = vec(0,1);
			eVec2(x,y,z,0) = vec(1,0);
			eVec2(x,y,z,1) = vec(1,1);		

			// ratio between different eigen values
			eRat(x,y,z,0) = val[1]*val[1]+val[0]*val[0] ;


			// calculate angles
			eAng(x,y,z,0) = acos( vec(0,1) );

		}
				
	}				

}


// Check dimensions of input slots and define parameters
template <typename T>
void DepthEdgeFeatures<T>::init(){
	
	if( this->filename() == "" )
		throw(std::runtime_error("DepthEdgeFeaturesWrite: empty path"));
	
	_maxEV_HSV.resize(this->numberOfScaleSpaces());
	_maxEVRatio_HSV.resize(this->numberOfScaleSpaces());
	_maxEV_RGB.resize(this->numberOfScaleSpaces());
	_maxEVRatio_RGB.resize(this->numberOfScaleSpaces());
	for( size_t i = 0; i < this->numberOfScaleSpaces(); i++ )
	{
		_maxEV_HSV[i].push_back(0);
		_maxEV_HSV[i].push_back(0);
		_maxEV_HSV[i].push_back(0);
		_maxEVRatio_HSV[i].push_back(0);
		_maxEVRatio_HSV[i].push_back(0);
		_maxEVRatio_HSV[i].push_back(0);
		_maxEV_RGB[i] = 0;
		_maxEVRatio_RGB[i] = 0;
	}
	
}


/// Returns an image containing the mean HSV colors of every pixel
CImg<float> getSegmentMeanColor(const CImg<float> &labels, const CImg<float> &img){
	
	CImg<float> labelSum(labels.max()+1,1,1,img.spectrum());
	std::vector<float> labelCount(labels.max()+1,0);
	cimg_forXYC(img,x,y,c)
	{
		labelSum(labels(x,y),0,0,c) += img(x,y,0,c);
		labelCount[labels(x,y)] ++;
	}
	cimg_forXYC(labelSum,x,y,c)
	{
		if(labelCount[x] != 0)
			labelSum(x,0,0,c) /= labelCount[x];
	}

	CImg<float> meanColor(img.width(), img.height(), 1, img.spectrum() );
	cimg_forXYC(meanColor,x,y,c)
	{
		meanColor(x,y,0,c) = labelSum( labels(x,y), 0, 0, c );
	}

	if(img.spectrum() == 3)
		meanColor.RGBtoHSV();
	return meanColor;
}


/// Returns the mean color of pixel neighbours in and against the given vector direction. Saved in z-channel
CImg<float> getNeighbourMeanColor( const CImg<float> &meanColor, const CImg<float> &eVec2 ){

	CImg<float> neighbourColor(meanColor.width(),meanColor.height(),2,meanColor.spectrum());
	cimg_forXYC(meanColor,x,y,c)
	{
		float color1 = meanColor.linear_atXYZC(float(x)+1.41*eVec2(x,y,0,0),float(y)+1.41*eVec2(x,y,0,1),0,c);
		float color2 = meanColor.linear_atXYZC(float(x)-1.41*eVec2(x,y,0,0),float(y)-1.41*eVec2(x,y,0,1),0,c);
		neighbourColor(x,y,0,c) = color1;
		neighbourColor(x,y,1,c) = color2;
	}
	return neighbourColor;

}



#endif /* _DEPTHEDGEFEATURES_HXX_ */
